"""Simple Q learning"""

import random, math

def exp_luce_choice(seq, mult = 1.0):
    '''Choose index of value in seq, treating value as probabilistic weight.'''
    exp_seq = [math.exp(x * mult) for x in seq]
    total = sum(exp_seq)
    if total:
        ran = random.random()
        scaled_total = 0.0
        for index, elem in enumerate(exp_seq):
            scaled_total += elem / total
            if ran < scaled_total:
                return index
        # In case we haven't found it (because of rounding errors?), return the last index
        return len(seq) - 1
    else:
        # All values are 0; pick a random position
        return random.randint(0, len(seq) - 1)

class Learner(object):

    gamma = .7
    """Discount rate"""
    eta = .1
    """Learning rate"""

    def __init__(self, states, actions, next_reinf=None, state1=None, gen_data=None):
        """Assign attributes and initialize the Q values."""
        self.states = states
        self.actions = actions
        self.q = dict([[s, [0.0 for a in range(len(actions))]] for s in states])
        self.next_reinf_func = next_reinf if next_reinf else lambda s, a, x: (self.get_random_state(), 0.0)
        self.gen_data = gen_data if gen_data else lambda: None
        self.state1 = state1

    def get_random_state(self):
        """Select a random state."""
        return random.choice(self.states)

    def get_best_q(self, state):
        """Get the best Q for a state."""
        return max(self.q[state])

    def get_best_action(self, state):
        """The action index with the highest Q for state."""
        values = self.q[state]
        best_value = values[0]
        best_action = 0
        for action in range(1, len(self.actions)):
            if values[action] > best_value:
                best_value = values[action]
                best_action = action
        return best_action

    def get_new_q(self, reinforcement, next_state):
        """New Q for a given reinforcement and next state."""
        return reinforcement + Learner.gamma * self.get_best_q(next_state)

    def learn(self, state, actionI, reinforcement, next_state):
        """Update the Q for state and action index, given a reinforcement and next state."""
        old_q = self.q[state][actionI]
        self.q[state][actionI] = \
                               (1.0 - Learner.eta) * old_q + \
                               Learner.eta * self.get_new_q(reinforcement, next_state)

    def decide(self, state, exploit=False):
        """Decide on an action (index) for a given state."""
        if exploit:
            return self.get_best_action(state)
        return exp_luce_choice(self.q[state], 2.0)

    def step(self, state, data, trace=False, learn=True, exploit=False):
        """Choose an action, find the reinforcement and next state, and learn."""
        actionI = self.decide(state, exploit=exploit)
        action = self.actions[actionI]
        next_state, reinf = self.next_reinf_func(state, action, data)
        if trace:
            print('State:%14s' % str(state), end='')
            if next_state: print('  action:%10s' % action, end='')
            if reinf: print('  reinf:%+5.1f' % reinf, end='')
            print()
        if next_state and learn:
            self.learn(state, actionI, reinf, next_state)
        return next_state

    def run1(self, trace=False):
        """Run the agent until its state is False."""
        state = self.state1 if self.state1 else random.choice(self.states)
        data = self.gen_data()
        if trace:
            print('Goal:', data)
        while state:
            state = self.step(state, data, trace=trace)

    def run(self, trials):
        """Run the agent trials times."""
        for t in range(trials):
            self.run1(trace=False)

    def test(self, exploit=True):
        """Test the agent through one run (until state is False)."""
        state = self.state1
        data = self.gen_data()
        print('Goal: %14s' % str(data))
        while state:
            state = self.step(state, data, trace=True, learn=False, exploit=exploit)

    def show_q(self):
        for state, qs in self.q.items():
            print(str(state).ljust(14), end='')
            for q in qs:
                print('%+.3f ' % q, end='')
            print()

